/**************************************************************
*    htron.c                                   	              *
*    H-Tron - a Tron motorcycle game                          *
*    Originally written in Borland Turbo C 2.01 for           *
*    XT-class DOS PCs equipped with a Hercules graphics card  *
*    Author: RobertK (RobertK@psc.at), December 2016          *
*    Version 2016-12-26a                                      *
*                                                             *
*    Ported to the Sinclair ZX81 (16K RAM)                    *
*    by RobertK in 2018 using the z88dk C compiler            *
*    ZX81 Version 2018-03-02                                  *
*    Thanks to A. Rea for providing ASM plot                  *
*    functions (zx81plot.h)                                   *
*                                                             *
*    This is the classic overhead view "Tron" game from       *
*    the movie of the same title (motorcycles that leave      *
*    "wall trails" behind them). If you crash into a wall,    *
*    your opponent scores a point. The player who first       *
*    reaches a score of 5 wins.                               *
*                                                             *
*    Controls:                                                *
*    Turn left/right:                                         *
*      Left Player: A/S                                       *
*      Right Player: Cursor left/right (5/8) or N/M           *
*      (in one-player mode, human is right player)            *
*    Pause game: P                                            *
*    End game: X                                              *
**************************************************************/

#include <stdio.h>		// required for zxTextOut()
#include <zx81plot.h>

#define POINTS_TO_WIN 5
#define AI_MAX_DISTANCE_TO_CHECK 12 // (DOS version: 15)

/* Playfield dimensions:  64 x 48  (DOS version:  90 x 58 )
   x=0 to 63 (DOS version:  89)
   y=0 to 47 (DOS version:  57)
*/

/* indicates whether the player has hit an obstacle */
int p1Crashed;
int p2Crashed;

/* player coordinate variables */
int p1x,p1y,p2x,p2y;

/* player direction variables in radian degrees */
/* 0 (or 360) = right, 90 = up, 180 = left, 270 = down */
int p1dir,p2dir;

/* player score variables */
int p1Score,p2Score;

/* One or two players */
int numberOfPlayers;

char *display;
extern int d_file @16396;


void cls()	// taken from z8048.c written by sirmorris
{
    // clears the screen
	int i;
	for (i = 0; i < 24; ++i)
	{
		memset(display + 33 * i, 0, 32);
	}
}


void zxTextOut(int x, int y, const char* text)	// taken from z8048.c written by sirmorris
{
	// writes a text to the specified x,y screen position
	char* dm = display + x + (33 * y);
	while (*text)
	{
		*dm = ascii_zx(*text);
		
		++text;
		++dm;
	}
}


void InitPlayfield()
{
  // sets up the initial playfield before a round starts
  int x,y;
  char textP1[25];
  char textP2[25];

  cls();

  /* Draw border */
  for (x = 0; x <= 63; ++x)
  {
	zx_plot(x,47,1);
	zx_plot(x,0,1);
  }
  for (y = 1; y <= 47; ++y)
  {
	zx_plot(0,y,1);
	zx_plot(63,y,1);
  }
  // score area separator  
  for (x = 0; x <= 63; ++x)
	zx_plot(x,5,1);
 

  /* display P1 and P2 score */
  // we first use sprintf() for string formatting,
  // and then we put the formatted string on the 
  // screen using zxTextOut()
  if (numberOfPlayers==0)
  {
   sprintf(textP1,"computer 1: %d",p1Score);
   sprintf(textP2,"computer 2: %d",p2Score);   
  }
  else if (numberOfPlayers==1)
  {
   sprintf(textP1,"computer: %d",p1Score);
   sprintf(textP2,"human: %d",p2Score);
  }
  else
  {
   sprintf(textP1,"player 1: %d",p1Score);
   sprintf(textP2,"player 2: %d",p2Score);
  }

  zxTextOut(1, 1, textP1);
  zxTextOut(32-strlen(textP2)-1, 1, textP2);
  
  /* Initiate player positions, directions and crash indicators */
  p1x=10;	// DOS version: 15
  p1y=26;	// DOS version: 29
  p2x=53;	// DOS version: 74
  p2y=26;	// DOS version: 29
  p1dir=0;
  p2dir=180;
  p1Crashed=0;
  p2Crashed=0;

  /* Draw players at their start positions */
  zx_plot(p1x,p1y,1);
  zx_plot(p2x,p2y,1);
}


void MovePlayers()
{
  // Moves both players one step forward
  switch(p1dir)
  {
    case 0:   p1x=p1x+1; break;
    case 90:  p1y=p1y-1; break;
    case 180: p1x=p1x-1; break;
    case 270: p1y=p1y+1; break;
  }
  switch(p2dir)
  {
    case 0:   p2x=p2x+1; break;
    case 90:  p2y=p2y-1; break;
    case 180: p2x=p2x-1; break;
    case 270: p2y=p2y+1; break;
  }

  /* Collision detection */
  if (zx_point(p1x,p1y)==1)
  {
    p1Crashed=1;
    p2Score++;
  }
  if (zx_point(p2x,p2y)==1)
  {
    p2Crashed=1;
    p1Score++;
  }
  /* no points when both players crashed */
  if (p1Crashed==1 && p2Crashed==1)
  {
   p1Score--;
   p2Score--;
  }

  /* Draw players at their new positions */
  /* (unless a collision has been detected) */
  if (p1Crashed==0) zx_plot(p1x,p1y,1);
  if (p2Crashed==0) zx_plot(p2x,p2y,1);

}


void AIComputerPlayer(int *x, int *y, int *dir)
{
  /* Let the computer player decide whether to turn left or right,
     or continue straight on.
     The coordinate and direction variables are passed by reference,
     so we can use this function for either computer player one or two */

  /* number of empty grid blocks before an obstacle
     from the AI player's point of view */
  int distanceFront,distanceLeft,distanceRight;

  int maxDistanceToChk;
  maxDistanceToChk=AI_MAX_DISTANCE_TO_CHECK;

  /* To make the AI player's movements less predictable,
  we occasionally make him a little short-sighted */
  maxDistanceToChk=maxDistanceToChk-(rand() % 8);	// DOS version: 10

  /* Check the distance from the AI player's position
     to the next obstacle in the front, left and right
     direction from the AI player's point of view. According
     to the players direction, we have to use the respective
     grid direction */
  switch(*dir)
  {
    case 0: /* right */
       distanceFront=AICheckDistanceEast(*x,*y,maxDistanceToChk);
       distanceLeft=AICheckDistanceNorth(*x,*y,maxDistanceToChk);
       distanceRight=AICheckDistanceSouth(*x,*y,maxDistanceToChk);
       break;
    case 90: /* up */
       distanceFront=AICheckDistanceNorth(*x,*y,maxDistanceToChk);
       distanceLeft=AICheckDistanceWest(*x,*y,maxDistanceToChk);
       distanceRight=AICheckDistanceEast(*x,*y,maxDistanceToChk);
       break;
    case 180: /* left */
       distanceFront=AICheckDistanceWest(*x,*y,maxDistanceToChk);
       distanceLeft=AICheckDistanceSouth(*x,*y,maxDistanceToChk);
       distanceRight=AICheckDistanceNorth(*x,*y,maxDistanceToChk);
       break;
    case 270: /* down */
       distanceFront=AICheckDistanceSouth(*x,*y,maxDistanceToChk);
       distanceLeft=AICheckDistanceEast(*x,*y,maxDistanceToChk);
       distanceRight=AICheckDistanceWest(*x,*y,maxDistanceToChk);
       break;
  }

  /* Now we decide what the AI player shall do */
  if ((distanceFront>=distanceLeft) && (distanceFront>=distanceRight))
  {
    /* clear sailing ahead, nothing to do */
  }
  else if ((distanceFront<distanceLeft) || (distanceFront<distanceRight))
  {
    /* now we know that it would be safer to turn either left or right */
    /* let us check which direction would be better */
    if (distanceLeft>distanceRight)
    {
    	/* turn left */
    	*dir = *dir+90;
    	if (*dir>=360) *dir=0;
    }
    else if (distanceLeft<distanceRight)
    {
        /* turn right */
		*dir = *dir-90;
		if (*dir<0) *dir=270;
    }
    else /* distanceLeft == distanceRight */
    {
		/* randomly turn either left or right */
		if ((rand() % 2)==0)
		{
		  /* turn left */
		  *dir = *dir+90;
		  if (*dir>=360) *dir=0;
		}
		else
		{
		  /* turn right */
		  *dir = *dir-90;
		  if (*dir<0) *dir=270;
		}
    }
  }
}


/* The following four functions calculate the distance from the
   given position to the next obstacle in grid map direction */
int AICheckDistanceEast(int x,int y,int maxDistanceToCheck)
{
  int distance;
  distance=1;
  while(zx_point(x+distance,y)==0 && distance<=maxDistanceToCheck)
  {
    distance++;
  }
  return distance;
}

int AICheckDistanceWest(int x,int y,int maxDistanceToCheck)
{
  int distance;
  distance=1;
  while(zx_point(x-distance,y)==0 && distance<=maxDistanceToCheck)
  {
    distance++;
  }
  return distance;
}

int AICheckDistanceNorth(int x,int y,int maxDistanceToCheck)
{
  int distance;
  distance=1;
  while(zx_point(x,y-distance)==0 && distance<=maxDistanceToCheck)
  {
    distance++;
  }
  return distance;
}

int AICheckDistanceSouth(int x,int y,int maxDistanceToCheck)
{
  int distance;
  distance=1;
  while(zx_point(x,y+distance)==0 && distance<=maxDistanceToCheck)
  {
    distance++;
  }
  return distance;
}


void main()
{
  int exitgame;
  int key;
  int notPlayAgain, restartAgain;
  
  // Key press indicators
  // Note: internally P1 is always the "right" player, i.e. human in 
  // 1-player-mode and player 2 in two player mode. 
  int p1PressedLeft;
  int p1PressedRight;
  int p2PressedLeft;
  int p2PressedRight;
  int lastState_p1PressedLeft;
  int lastState_p1PressedRight;
  int lastState_p2PressedLeft;
  int lastState_p2PressedRight;
  
  restartAgain=0;
  randomize();

  display = (char *) d_file+1;

  do
  {
  
	  notPlayAgain=0;
	  cls();
	  printf("\x0c*** h-tron ***\n");  // the "\x0c" (0x0c character) at the beginning resets the cursor position to the top left corner 
	  printf("by robertk (2016/2018)\n\n");
	  printf("controls:\n\n");
	  printf("turn left/right:\n");
	  printf(" left player: a/s\n");
	  printf(" right player: cursor left/right");
	  printf(" (5/8) or n/m\n");
	  printf(" (in one-player mode,\n");
	  printf("  human is right player)\n\n");
	  printf("pause game: p, end game: x\n\n");
	  printf("press space or 0\n");
	  printf("to start a round\n\n");
	  printf("one or two players (1/2)?\n");
	  printf("(or press d for demo mode)");

	  // printf("in_lookupkey() for n is %d\n",in_LookupKey('N'));	  

	  do
	  {
		key=in_Inkey();
		
		// if (key>0) printf("\nkeycode: %d",key);				
		// printf("key a (65) pressed: %d\n",in_KeyPressed(in_LookupKey('A')));
		
		if (key==68) numberOfPlayers=0; // d
		if (key==49) numberOfPlayers=1; // 1
		if (key==50) numberOfPlayers=2; // 2
		if (key==88) // x
		{
		  cls();
		  return;
		}
	  }
	  while(key!=68 && key!=49 && key!=50); // d, 1 or 2


	  /* game restart loop */
	  while(notPlayAgain<1)
	  {
		p1Score=0;
		p2Score=0;
		exitgame=0;

		/* round restart loop */
		while(p1Score<POINTS_TO_WIN && p2Score<POINTS_TO_WIN && exitgame<1)
		{
			InitPlayfield();

			if (numberOfPlayers>0)
			{
			  do
			  {
				key=in_Inkey();
				if (key==88) exitgame=1;	// x pressed
			  } while (key!=32 && key!= 48 && key!=88); // space, 0, x
			}
			else in_Pause(1000);  /* in demo mode, the round starts automatically */

			lastState_p1PressedLeft=0;
			lastState_p1PressedRight=0;
			lastState_p2PressedLeft=0;
			lastState_p2PressedRight=0;
			
			/* main game loop */
			while(exitgame<1 && p1Crashed<1 && p2Crashed<1)
			{						
				// Check keyboard state for player's control
				
				// Here we could use in_Inkey() and then switch-case to set the 
				// actions for the possible key combinations. But this would not allow 
				// both players to simultaneously press their control key - one would lock 
				// the other one's steering.
				// So we have to use in_KeyPressed() which requires more code lines and 
				// is probably a little less efficient, but it does the job that we need.
				
				/* Lookup keys, as determined with the in_LookupKey() function:
				5: 4343
				N: 2175
				Shift-5: not required (the cursor keys in the EightyOne emulator trigger the ZX81's 5 and 8 keys)
				--
				8: 1263
				M: 1151
				Shift-8: not required, see above
				--
				A: 509
				S: 765
				*/
				
				p1PressedLeft=0;
				p1PressedRight=0;
				p2PressedLeft=0;
				p2PressedRight=0;				
				if (in_KeyPressed(4343)>0 || in_KeyPressed(2175)>0) p1PressedLeft=1;  // 5 or n
				if (in_KeyPressed(1263)>0 || in_KeyPressed(1151)>0) p1PressedRight=1; // 8 or m
				if (in_KeyPressed(509)>0) p2PressedLeft=1;  // a
				if (in_KeyPressed(765)>0) p2PressedRight=1; // s

				// We react to the key press only if there was a change since the 
				// last time we came along here, hence the lastState...-Check. 
				// By doing this we avoid unintended multiple turn commands, 
				// the player must always first release the button and then press it again.
				
				if (p1PressedLeft>0 && p1PressedLeft!=lastState_p1PressedLeft)
				{
				  if (numberOfPlayers>0)
				  {
					p2dir = p2dir+90;
					if (p2dir>=360) p2dir=0;
				  }
				}
				
				if (p1PressedRight>0 && p1PressedRight!=lastState_p1PressedRight)
				{
				  if (numberOfPlayers>0)
				  {
					p2dir = p2dir-90;
					if (p2dir<0) p2dir=270;
				  }
				}

				if (p2PressedLeft>0 && p2PressedLeft!=lastState_p2PressedLeft)
				{
				  if (numberOfPlayers==2)
				  {
					p1dir = p1dir+90;
					if (p1dir>=360) p1dir=0;
				  }
				}

				if (p2PressedRight>0 && p2PressedRight!=lastState_p2PressedRight)
				{
				  if (numberOfPlayers==2)
				  {
					p1dir = p1dir-90;
					if (p1dir<0) p1dir=270;
				  }
				 }
			
			
				// The remaining keys we can check with a simple in_Key() call
				key=in_Inkey();
				switch(key)
				{
				case 80: /* P Key (Pause) */	// 25
					 in_Wait(50);  	// sleep for 50 ms before waiting for new key
					 in_WaitForKey();
					 break;
				case 88: /* x key */		// 1
					exitgame=1;
					break;
				}


				/* let the computer player decide what to do */
				/* in one-player mode, Player 1 is controlled by the computer */
				if (numberOfPlayers<2) AIComputerPlayer(&p1x,&p1y,&p1dir);
				/* in demo mode, Player 2 is also controlled by the computer */
				if (numberOfPlayers==0) AIComputerPlayer(&p2x,&p2y,&p2dir);

				MovePlayers();

				// speed throttling
				// depending on the type of game we use different delay values
				switch(numberOfPlayers)
				{
					case 0: break;	// no delay required because AI for two computer players is most demanding for the CPU
					case 1: in_Wait(8); break;  // Sleep for 8 milliseconds
					case 2: in_Wait(18); break;  // Sleep for 18 milliseconds
				}		  
				  
				// remember the last state of the control keys			  
				lastState_p1PressedLeft=p1PressedLeft;
				lastState_p1PressedRight=p1PressedRight;
				lastState_p2PressedLeft=p2PressedLeft;
				lastState_p2PressedRight=p2PressedRight;			  
			
			}

			/* wait a quarter of a second after a crash has
			   happened before clearing the screen */
			if (p1Crashed || p2Crashed) in_Wait(250);
		
		}

		cls();
		printf("\x0cfinal score    %d : %d\n\n", p1Score, p2Score);  // the "\x0c" (0x0c character) at the beginning resets the cursor position to the top left corner
		
		if (p1Score>p2Score)
			if (numberOfPlayers==0)
			  printf("computer player one wins\n");
			else if (numberOfPlayers==2)
			  printf("player one wins\n");
			else
			  printf("computer wins\n");
		if (p1Score<p2Score)
			if (numberOfPlayers==0)
			  printf("computer player two wins\n");
			else if (numberOfPlayers==2)
			  printf("player two wins\n");
			else
			  printf("you win\n");
		if (p1Score==p2Score)
			printf("a draw\n");
			
		printf("\nplay again (y/n)?\n\n");
		printf("(press m to return to the menu)\n");
		do
		{
			key=in_Inkey();
			if (key==78) // n
			{
				notPlayAgain=1;
				restartAgain=0;	
			}
			if (key==77) // m
			{
				notPlayAgain=1;
				restartAgain=1;	
			}

		} while(key!=89 && key!=78 && key!=48 && key!=77);  // y or n or 0 or m, 0 is for the joystick fire button

	  }
  } while (restartAgain>0);
  
  cls();
  return;
}
